package fit;

import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;
import java.util.Vector;

public abstract class Invocation {
    protected StepFixture targetObject;
    private Parse cells;
    private Method targetMethod;
    private Vector argCells;  // it's convenient to refer to cells by index.


    Invocation(StepFixture targetObject, Parse cells) throws NoSuchMethodException {
        this.targetObject = targetObject;
        this.cells = cells;
        argCells = new Vector();
        for (Parse args = this.cells.more;
             args != null && !args.text().equals("");
             args = args.more) {
            argCells.add(args);
        }
    }

    /// Plain and lazy accessors //////////////////////////

    public Vector argCells() {
        return argCells;
    }

    protected Parse cells() {
        return cells;
    }

    // Method is lazily gotten because an exception thrown
     // in a constructor causes FIT to annotate the previous
     // row, not the row currently being evaluated.
     public Method targetMethod() throws NoSuchMethodException {
         if (targetMethod == null) {
             targetMethod = findTargetMethod();
         }
         return targetMethod;
     }


    /// Other accessor-like things //////////////////////

    public String stepName() {
        return cells.text();
    }

    public Parse lastArgCell() {
        return (Parse) argCells.lastElement();
    }

    protected int numberOfGivenArgs() {
            return argCells.size();
        }

    public Parse argCellAt(int index) {
        return (Parse) argCells.elementAt(index);
    }



    /// The real action. ////////////
    abstract protected Method findTargetMethod() throws NoSuchMethodException;
    abstract void execute() throws Exception, InvocationTargetException;


    protected Method findMethodByArity(Object onWhere, String name, int arity) throws NoSuchMethodException {
        Method[] methods = onWhere.getClass().getMethods();
        for (int i = 0; i < methods.length; i++) {
            Method candidate = methods[i];
            if (candidate.getName().equals(Fixture.camel(name)) &&
                candidate.getParameterTypes().length == arity) {
                    return candidate;
            }
        }
       throw new NoSuchMethodException();
    }


    protected Object[] argumentValues() throws Exception {
        return argumentValuesInRange(targetMethod(), 0, numberOfGivenArgs());
    }

    protected Object[] argumentValuesInRange(Method method, int inclusiveStart, int exclusiveEnd) throws Exception {
        Object[] retval = new Object[exclusiveEnd-inclusiveStart];

        Class[] types = method.getParameterTypes();
        for (int rangeIndex = inclusiveStart; rangeIndex < exclusiveEnd; rangeIndex++) {
            int arrayIndex = rangeIndex - inclusiveStart;
            TypeAdapter typeAdapter = TypeAdapter.on(targetObject, types[arrayIndex]);
            retval[arrayIndex] = typeAdapter.parse(argCellAt(rangeIndex).text());
        }
        return retval;
    }

}

